@@ -134,7 +134,7 @@ def group_assignment(tlist):
         return token.match(T.Assignment, ':=')
 
     def valid(token):
-        return token is not None
+        return token is not None and token.ttype not in (T.Keyword)
 
     def post(tlist, pidx, tidx, nidx):
         m_semicolon = T.Punctuation, ';'
@@ -274,7 +274,7 @@ def group_where(tlist):
 @recurse()
 def group_aliased(tlist):
     I_ALIAS = (sql.Parenthesis, sql.Function, sql.Case, sql.Identifier,
-               sql.Operation)
+               sql.Operation, sql.Comparison)
 
     tidx, token = tlist.token_next_by(i=I_ALIAS, t=T.Number)
     while token:
@@ -346,10 +346,10 @@ def group(stmt):
         group_order,
         group_typecasts,
         group_operator,
+        group_comparison,
         group_as,
         group_aliased,
         group_assignment,
-        group_comparison,
 
         align_comments,
         group_identifier_list,
@@ -380,7 +380,7 @@ def _group(tlist, cls, match,
 
         if match(token):
             nidx, next_ = tlist.token_next(tidx)
-            if valid_prev(prev_) and valid_next(next_):
+            if prev_ and valid_prev(prev_) and valid_next(next_):
                 from_idx, to_idx = post(tlist, pidx, tidx, nidx)
                 grp = tlist.group_tokens(cls, from_idx, to_idx, extend=extend)
 
