@@ -1,13 +1,21 @@
 # -*- coding: utf-8 -*-
+#
+# Copyright (C) 2016 Andi Albrecht, albrecht.andi@gmail.com
+#
+# This module is part of python-sqlparse and is released under
+# the BSD License: https://opensource.org/licenses/BSD-3-Clause
 
 """This module contains classes representing syntactical elements of SQL."""
+from __future__ import print_function
 
 import re
-import sys
 
 from sqlparse import tokens as T
+from sqlparse.compat import string_types, text_type, unicode_compatible
+from sqlparse.utils import imt, remove_quotes
 
 
+@unicode_compatible
 class Token(object):
     """Base class for all other classes in this module.
 
@@ -16,51 +24,42 @@ class Token(object):
     the type of the token.
     """
 
-    __slots__ = ('value', 'ttype', 'parent', 'normalized', 'is_keyword')
+    __slots__ = ('value', 'ttype', 'parent', 'normalized', 'is_keyword',
+                 'is_group', 'is_whitespace')
 
     def __init__(self, ttype, value):
+        value = text_type(value)
         self.value = value
-        if ttype in T.Keyword:
-            self.normalized = value.upper()
-        else:
-            self.normalized = value
         self.ttype = ttype
-        self.is_keyword = ttype in T.Keyword
         self.parent = None
+        self.is_group = False
+        self.is_keyword = ttype in T.Keyword
+        self.is_whitespace = self.ttype in T.Whitespace
+        self.normalized = value.upper() if self.is_keyword else value
 
     def __str__(self):
-        if sys.version_info[0] == 3:
-            return self.value
-        else:
-            return unicode(self).encode('utf-8')
+        return self.value
+
+    # Pending tokenlist __len__ bug fix
+    # def __len__(self):
+    #     return len(self.value)
 
     def __repr__(self):
-        short = self._get_repr_value()
-        if sys.version_info[0] < 3:
-            short = short.encode('utf-8')
-        return '<%s \'%s\' at 0x%07x>' % (self._get_repr_name(),
-                                          short, id(self))
+        cls = self._get_repr_name()
+        value = self._get_repr_value()
 
-    def __unicode__(self):
-        """Returns a unicode representation of this object."""
-        return self.value or ''
-
-    def to_unicode(self):
-        """Returns a unicode representation of this object.
-
-        .. deprecated:: 0.1.5
-           Use ``unicode(token)`` (for Python 3: ``str(token)``) instead.
-        """
-        return unicode(self)
+        q = u'"' if value.startswith("'") and value.endswith("'") else u"'"
+        return u"<{cls} {q}{value}{q} at 0x{id:2X}>".format(
+            id=id(self), **locals())
 
     def _get_repr_name(self):
         return str(self.ttype).split('.')[-1]
 
     def _get_repr_value(self):
-        raw = unicode(self)
+        raw = text_type(self)
         if len(raw) > 7:
-            raw = raw[:6] + u'...'
-        return re.sub('\s+', ' ', raw)
+            raw = raw[:6] + '...'
+        return re.sub(r'\s+', ' ', raw)
 
     def flatten(self):
         """Resolve subgroups."""
@@ -82,40 +81,23 @@ class Token(object):
         if not type_matched or values is None:
             return type_matched
 
-        if regex:
-            if isinstance(values, basestring):
-                values = set([values])
+        if isinstance(values, string_types):
+            values = (values,)
 
-            if self.ttype is T.Keyword:
-                values = set(re.compile(v, re.IGNORECASE) for v in values)
-            else:
-                values = set(re.compile(v) for v in values)
+        if regex:
+            # TODO: Add test for regex with is_keyboard = false
+            flag = re.IGNORECASE if self.is_keyword else 0
+            values = (re.compile(v, flag) for v in values)
 
             for pattern in values:
-                if pattern.search(self.value):
+                if pattern.search(self.normalized):
                     return True
             return False
 
-        if isinstance(values, basestring):
-            if self.is_keyword:
-                return values.upper() == self.normalized
-            return values == self.value
-
         if self.is_keyword:
-            for v in values:
-                if v.upper() == self.normalized:
-                    return True
-            return False
+            values = (v.upper() for v in values)
 
-        return self.value in values
-
-    def is_group(self):
-        """Returns ``True`` if this object has children."""
-        return False
-
-    def is_whitespace(self):
-        """Return ``True`` if this token is a whitespace token."""
-        return self.ttype and self.ttype in T.Whitespace
+        return self.normalized in values
 
     def within(self, group_cls):
         """Returns ``True`` if this token is within *group_cls*.
@@ -144,6 +126,7 @@ class Token(object):
         return False
 
 
+@unicode_compatible
 class TokenList(Token):
     """A group of tokens.
 
@@ -151,60 +134,50 @@ class TokenList(Token):
     list of child-tokens.
     """
 
-    __slots__ = ('value', 'ttype', 'tokens')
+    __slots__ = 'tokens'
 
     def __init__(self, tokens=None):
-        if tokens is None:
-            tokens = []
-        self.tokens = tokens
-        Token.__init__(self, None, self._to_string())
-
-    def __unicode__(self):
-        return self._to_string()
+        self.tokens = tokens or []
+        [setattr(token, 'parent', self) for token in tokens]
+        super(TokenList, self).__init__(None, text_type(self))
+        self.is_group = True
 
     def __str__(self):
-        str_ = self._to_string()
-        if sys.version_info[0] < 2:
-            str_ = str_.encode('utf-8')
-        return str_
+        return u''.join(token.value for token in self.flatten())
 
-    def _to_string(self):
-        if sys.version_info[0] == 3:
-            return ''.join(x.value for x in self.flatten())
-        else:
-            return ''.join(unicode(x) for x in self.flatten())
+    # weird bug
+    # def __len__(self):
+    #     return len(self.tokens)
+
+    def __iter__(self):
+        return iter(self.tokens)
+
+    def __getitem__(self, item):
+        return self.tokens[item]
 
     def _get_repr_name(self):
-        return self.__class__.__name__
+        return type(self).__name__
 
-    def _pprint_tree(self, max_depth=None, depth=0):
+    def _pprint_tree(self, max_depth=None, depth=0, f=None):
         """Pretty-print the object tree."""
-        indent = ' ' * (depth * 2)
+        indent = u' | ' * depth
         for idx, token in enumerate(self.tokens):
-            if token.is_group():
-                pre = ' +-'
-            else:
-                pre = ' | '
-            print '%s%s%d %s \'%s\'' % (indent, pre, idx,
-                                        token._get_repr_name(),
-                                        token._get_repr_value())
-            if (token.is_group() and (max_depth is None or depth < max_depth)):
-                token._pprint_tree(max_depth, depth + 1)
+            cls = token._get_repr_name()
+            value = token._get_repr_value()
 
-    def _remove_quotes(self, val):
-        """Helper that removes surrounding quotes from strings."""
-        if not val:
-            return val
-        if val[0] in ('"', '\'') and val[-1] == val[0]:
-            val = val[1:-1]
-        return val
+            q = u'"' if value.startswith("'") and value.endswith("'") else u"'"
+            print(u"{indent}{idx:2d} {cls} {q}{value}{q}"
+                  .format(**locals()), file=f)
+
+            if token.is_group and (max_depth is None or depth < max_depth):
+                token._pprint_tree(max_depth, depth + 1, f)
 
     def get_token_at_offset(self, offset):
         """Returns the token that is on position offset."""
         idx = 0
         for token in self.flatten():
             end = idx + len(token.value)
-            if idx <= offset <= end:
+            if idx <= offset < end:
                 return token
             idx = end
 
@@ -214,188 +187,146 @@ class TokenList(Token):
         This method is recursively called for all child tokens.
         """
         for token in self.tokens:
-            if isinstance(token, TokenList):
+            if token.is_group:
                 for item in token.flatten():
                     yield item
             else:
                 yield token
 
-#    def __iter__(self):
-#        return self
-#
-#    def next(self):
-#        for token in self.tokens:
-#            yield token
-
-    def is_group(self):
-        return True
-
     def get_sublists(self):
-#        return [x for x in self.tokens if isinstance(x, TokenList)]
-        for x in self.tokens:
-            if isinstance(x, TokenList):
-                yield x
+        for token in self.tokens:
+            if token.is_group:
+                yield token
 
     @property
     def _groupable_tokens(self):
         return self.tokens
 
-    def token_first(self, ignore_whitespace=True, ignore_comments=False):
+    def _token_matching(self, funcs, start=0, end=None, reverse=False):
+        """next token that match functions"""
+        if start is None:
+            return None
+
+        if not isinstance(funcs, (list, tuple)):
+            funcs = (funcs,)
+
+        if reverse:
+            assert end is None
+            for idx in range(start - 2, -1, -1):
+                token = self.tokens[idx]
+                for func in funcs:
+                    if func(token):
+                        return idx, token
+        else:
+            for idx, token in enumerate(self.tokens[start:end], start=start):
+                for func in funcs:
+                    if func(token):
+                        return idx, token
+        return None, None
+
+    def token_first(self, skip_ws=True, skip_cm=False):
         """Returns the first child token.
 
-        If *ignore_whitespace* is ``True`` (the default), whitespace
+        If *skip_ws* is ``True`` (the default), whitespace
         tokens are ignored.
 
-        if *ignore_comments* is ``True`` (default: ``False``), comments are
+        if *skip_cm* is ``True`` (default: ``False``), comments are
         ignored too.
         """
-        for token in self.tokens:
-            if ignore_whitespace and token.is_whitespace():
-                continue
-            if ignore_comments and isinstance(token, Comment):
-                continue
-            return token
+        # this on is inconsistent, using Comment instead of T.Comment...
+        funcs = lambda tk: not ((skip_ws and tk.is_whitespace) or
+                                (skip_cm and imt(tk, t=T.Comment, i=Comment)))
+        return self._token_matching(funcs)[1]
 
-    def token_next_by_instance(self, idx, clss, end=None):
-        """Returns the next token matching a class.
+    def token_next_by(self, i=None, m=None, t=None, idx=-1, end=None):
+        funcs = lambda tk: imt(tk, i, m, t)
+        idx += 1
+        return self._token_matching(funcs, idx, end)
 
-        *idx* is where to start searching in the list of child tokens.
-        *clss* is a list of classes the token should be an instance of.
+    def token_not_matching(self, funcs, idx):
+        funcs = (funcs,) if not isinstance(funcs, (list, tuple)) else funcs
+        funcs = [lambda tk: not func(tk) for func in funcs]
+        return self._token_matching(funcs, idx)
 
-        If no matching token can be found ``None`` is returned.
-        """
-        if not isinstance(clss, (list, tuple)):
-            clss = (clss,)
+    def token_matching(self, funcs, idx):
+        return self._token_matching(funcs, idx)[1]
 
-        for token in self.tokens[idx:end]:
-            if isinstance(token, clss):
-                return token
-
-    def token_next_by_type(self, idx, ttypes):
-        """Returns next matching token by it's token type."""
-        if not isinstance(ttypes, (list, tuple)):
-            ttypes = [ttypes]
-
-        for token in self.tokens[idx:]:
-            if token.ttype in ttypes:
-                return token
-
-    def token_next_match(self, idx, ttype, value, regex=False):
-        """Returns next token where it's ``match`` method returns ``True``."""
-        if not isinstance(idx, int):
-            idx = self.token_index(idx)
-
-        for n in xrange(idx, len(self.tokens)):
-            token = self.tokens[n]
-            if token.match(ttype, value, regex):
-                return token
-
-    def token_not_matching(self, idx, funcs):
-        for token in self.tokens[idx:]:
-            passed = False
-            for func in funcs:
-                if func(token):
-                    passed = True
-                    break
-
-            if not passed:
-                return token
-
-    def token_matching(self, idx, funcs):
-        for token in self.tokens[idx:]:
-            for func in funcs:
-                if func(token):
-                    return token
-
-    def token_prev(self, idx, skip_ws=True):
+    def token_prev(self, idx, skip_ws=True, skip_cm=False):
         """Returns the previous token relative to *idx*.
 
         If *skip_ws* is ``True`` (the default) whitespace tokens are ignored.
+        If *skip_cm* is ``True`` comments are ignored.
         ``None`` is returned if there's no previous token.
         """
-        if idx is None:
-            return None
+        return self.token_next(idx, skip_ws, skip_cm, _reverse=True)
 
-        if not isinstance(idx, int):
-            idx = self.token_index(idx)
-
-        while idx:
-            idx -= 1
-            if self.tokens[idx].is_whitespace() and skip_ws:
-                continue
-            return self.tokens[idx]
-
-    def token_next(self, idx, skip_ws=True):
+    # TODO: May need to re-add default value to idx
+    def token_next(self, idx, skip_ws=True, skip_cm=False, _reverse=False):
         """Returns the next token relative to *idx*.
 
         If *skip_ws* is ``True`` (the default) whitespace tokens are ignored.
+        If *skip_cm* is ``True`` comments are ignored.
         ``None`` is returned if there's no next token.
         """
         if idx is None:
-            return None
-
-        if not isinstance(idx, int):
-            idx = self.token_index(idx)
-
-        while idx < len(self.tokens) - 1:
-            idx += 1
-            if self.tokens[idx].is_whitespace() and skip_ws:
-                continue
-            return self.tokens[idx]
+            return None, None
+        idx += 1  # alot of code usage current pre-compensates for this
+        funcs = lambda tk: not ((skip_ws and tk.is_whitespace) or
+                                (skip_cm and imt(tk, t=T.Comment, i=Comment)))
+        return self._token_matching(funcs, idx, reverse=_reverse)
 
     def token_index(self, token, start=0):
         """Return list index of token."""
-        if start > 0:
-            # Performing `index` manually is much faster when starting in the middle
-            # of the list of tokens and expecting to find the token near to the starting
-            # index.
-            for i in xrange(start, len(self.tokens)):
-                if self.tokens[i] == token:
-                    return i
-            return -1
-        return self.tokens.index(token)
+        start = start if isinstance(start, int) else self.token_index(start)
+        return start + self.tokens[start:].index(token)
 
-    def tokens_between(self, start, end, exclude_end=False):
-        """Return all tokens between (and including) start and end.
-
-        If *exclude_end* is ``True`` (default is ``False``) the end token
-        is included too.
-        """
-        # FIXME(andi): rename exclude_end to inlcude_end
-        if exclude_end:
-            offset = 0
-        else:
-            offset = 1
-        end_idx = self.token_index(end) + offset
-        start_idx = self.token_index(start)
-        return self.tokens[start_idx:end_idx]
-
-    def group_tokens(self, grp_cls, tokens, ignore_ws=False):
+    def group_tokens(self, grp_cls, start, end, include_end=True,
+                     extend=False):
         """Replace tokens by an instance of *grp_cls*."""
-        idx = self.token_index(tokens[0])
-        if ignore_ws:
-            while tokens and tokens[-1].is_whitespace():
-                tokens = tokens[:-1]
-        for t in tokens:
-            self.tokens.remove(t)
-        grp = grp_cls(tokens)
-        for token in tokens:
+        start_idx = start
+        start = self.tokens[start_idx]
+
+        end_idx = end + include_end
+
+        # will be needed later for new group_clauses
+        # while skip_ws and tokens and tokens[-1].is_whitespace:
+        #     tokens = tokens[:-1]
+
+        if extend and isinstance(start, grp_cls):
+            subtokens = self.tokens[start_idx + 1:end_idx]
+
+            grp = start
+            grp.tokens.extend(subtokens)
+            del self.tokens[start_idx + 1:end_idx]
+            grp.value = text_type(start)
+        else:
+            subtokens = self.tokens[start_idx:end_idx]
+            grp = grp_cls(subtokens)
+            self.tokens[start_idx:end_idx] = [grp]
+            grp.parent = self
+
+        for token in subtokens:
             token.parent = grp
-        grp.parent = self
-        self.tokens.insert(idx, grp)
+
         return grp
 
     def insert_before(self, where, token):
         """Inserts *token* before *where*."""
-        self.tokens.insert(self.token_index(where), token)
+        if not isinstance(where, int):
+            where = self.token_index(where)
+        token.parent = self
+        self.tokens.insert(where, token)
 
     def insert_after(self, where, token, skip_ws=True):
         """Inserts *token* after *where*."""
-        next_token = self.token_next(where, skip_ws=skip_ws)
-        if next_token is None:
+        if not isinstance(where, int):
+            where = self.token_index(where)
+        nidx, next_ = self.token_next(where, skip_ws=skip_ws)
+        token.parent = self
+        if next_ is None:
             self.tokens.append(token)
         else:
-            self.tokens.insert(self.token_index(next_token), token)
+            self.tokens.insert(nidx, token)
 
     def has_alias(self):
         """Returns ``True`` if an alias is present."""
@@ -405,17 +336,15 @@ class TokenList(Token):
         """Returns the alias for this identifier or ``None``."""
 
         # "name AS alias"
-        kw = self.token_next_match(0, T.Keyword, 'AS')
+        kw_idx, kw = self.token_next_by(m=(T.Keyword, 'AS'))
         if kw is not None:
-            return self._get_first_name(kw, keywords=True)
+            return self._get_first_name(kw_idx + 1, keywords=True)
 
         # "name alias" or "complicated column expression alias"
-        if len(self.tokens) > 2 \
-           and self.token_next_by_type(0, T.Whitespace) is not None:
+        _, ws = self.token_next_by(t=T.Whitespace)
+        if len(self.tokens) > 2 and ws is not None:
             return self._get_first_name(reverse=True)
 
-        return None
-
     def get_name(self):
         """Returns the name of this identifier.
 
@@ -423,39 +352,26 @@ class TokenList(Token):
         be considered as the name under which the object corresponding to
         this identifier is known within the current statement.
         """
-        alias = self.get_alias()
-        if alias is not None:
-            return alias
-        return self.get_real_name()
+        return self.get_alias() or self.get_real_name()
 
     def get_real_name(self):
         """Returns the real name (object name) of this identifier."""
         # a.b
-        dot = self.token_next_match(0, T.Punctuation, '.')
-        if dot is not None:
-            return self._get_first_name(self.token_index(dot))
-
-        return self._get_first_name()
+        dot_idx, _ = self.token_next_by(m=(T.Punctuation, '.'))
+        return self._get_first_name(dot_idx)
 
     def get_parent_name(self):
         """Return name of the parent object if any.
 
         A parent object is identified by the first occuring dot.
         """
-        dot = self.token_next_match(0, T.Punctuation, '.')
-        if dot is None:
-            return None
-        prev_ = self.token_prev(self.token_index(dot))
-        if prev_ is None:  # something must be verry wrong here..
-            return None
-        return self._remove_quotes(prev_.value)
+        dot_idx, _ = self.token_next_by(m=(T.Punctuation, '.'))
+        _, prev_ = self.token_prev(dot_idx)
+        return remove_quotes(prev_.value) if prev_ is not None else None
 
     def _get_first_name(self, idx=None, reverse=False, keywords=False):
         """Returns the name of the first token with a name"""
 
-        if idx and not isinstance(idx, int):
-            idx = self.token_index(idx) + 1
-
         tokens = self.tokens[idx:] if idx else self.tokens
         tokens = reversed(tokens) if reverse else tokens
         types = [T.Name, T.Wildcard, T.String.Symbol]
@@ -463,18 +379,16 @@ class TokenList(Token):
         if keywords:
             types.append(T.Keyword)
 
-        for tok in tokens:
-            if tok.ttype in types:
-                return self._remove_quotes(tok.value)
-            elif isinstance(tok, Identifier) or isinstance(tok, Function):
-                return tok.get_name()
-        return None
+        for token in tokens:
+            if token.ttype in types:
+                return remove_quotes(token.value)
+            elif isinstance(token, (Identifier, Function)):
+                return token.get_name()
+
 
 class Statement(TokenList):
     """Represents a SQL statement."""
 
-    __slots__ = ('value', 'ttype', 'tokens')
-
     def get_type(self):
         """Returns the type of a statement.
 
@@ -485,7 +399,7 @@ class Statement(TokenList):
         Whitespaces and comments at the beginning of the statement
         are ignored.
         """
-        first_token = self.token_first(ignore_comments=True)
+        first_token = self.token_first(skip_cm=True)
         if first_token is None:
             # An "empty" statement that either has not tokens at all
             # or only whitespace tokens.
@@ -494,6 +408,19 @@ class Statement(TokenList):
         elif first_token.ttype in (T.Keyword.DML, T.Keyword.DDL):
             return first_token.normalized
 
+        elif first_token.ttype == T.Keyword.CTE:
+            # The WITH keyword should be followed by either an Identifier or
+            # an IdentifierList containing the CTE definitions;  the actual
+            # DML keyword (e.g. SELECT, INSERT) will follow next.
+            fidx = self.token_index(first_token)
+            tidx, token = self.token_next(fidx, skip_ws=True)
+            if isinstance(token, (Identifier, IdentifierList)):
+                _, dml_keyword = self.token_next(tidx, skip_ws=True)
+
+                if dml_keyword.ttype == T.Keyword.DML:
+                    return dml_keyword.normalized
+
+        # Hmm, probably invalid syntax, so return unknown.
         return 'UNKNOWN'
 
 
@@ -503,57 +430,48 @@ class Identifier(TokenList):
     Identifiers may have aliases or typecasts.
     """
 
-    __slots__ = ('value', 'ttype', 'tokens')
-
     def is_wildcard(self):
         """Return ``True`` if this identifier contains a wildcard."""
-        token = self.token_next_by_type(0, T.Wildcard)
+        _, token = self.token_next_by(t=T.Wildcard)
         return token is not None
 
     def get_typecast(self):
         """Returns the typecast or ``None`` of this object as a string."""
-        marker = self.token_next_match(0, T.Punctuation, '::')
-        if marker is None:
-            return None
-        next_ = self.token_next(self.token_index(marker), False)
-        if next_ is None:
-            return None
-        return unicode(next_)
+        midx, marker = self.token_next_by(m=(T.Punctuation, '::'))
+        nidx, next_ = self.token_next(midx, skip_ws=False)
+        return next_.value if next_ else None
 
     def get_ordering(self):
         """Returns the ordering or ``None`` as uppercase string."""
-        ordering = self.token_next_by_type(0, T.Keyword.Order)
-        if ordering is None:
-            return None
-        return ordering.value.upper()
+        _, ordering = self.token_next_by(t=T.Keyword.Order)
+        return ordering.normalized if ordering else None
 
     def get_array_indices(self):
         """Returns an iterator of index token lists"""
 
-        for tok in self.tokens:
-            if isinstance(tok, SquareBrackets):
+        for token in self.tokens:
+            if isinstance(token, SquareBrackets):
                 # Use [1:-1] index to discard the square brackets
-                yield tok.tokens[1:-1]
+                yield token.tokens[1:-1]
 
 
 class IdentifierList(TokenList):
     """A list of :class:`~sqlparse.sql.Identifier`\'s."""
 
-    __slots__ = ('value', 'ttype', 'tokens')
-
     def get_identifiers(self):
         """Returns the identifiers.
 
         Whitespaces and punctuations are not included in this generator.
         """
-        for x in self.tokens:
-            if not x.is_whitespace() and not x.match(T.Punctuation, ','):
-                yield x
+        for token in self.tokens:
+            if not (token.is_whitespace or token.match(T.Punctuation, ',')):
+                yield token
 
 
 class Parenthesis(TokenList):
     """Tokens between parenthesis."""
-    __slots__ = ('value', 'ttype', 'tokens')
+    M_OPEN = T.Punctuation, '('
+    M_CLOSE = T.Punctuation, ')'
 
     @property
     def _groupable_tokens(self):
@@ -562,31 +480,32 @@ class Parenthesis(TokenList):
 
 class SquareBrackets(TokenList):
     """Tokens between square brackets"""
-
-    __slots__ = ('value', 'ttype', 'tokens')
+    M_OPEN = T.Punctuation, '['
+    M_CLOSE = T.Punctuation, ']'
 
     @property
     def _groupable_tokens(self):
         return self.tokens[1:-1]
 
+
 class Assignment(TokenList):
     """An assignment like 'var := val;'"""
-    __slots__ = ('value', 'ttype', 'tokens')
 
 
 class If(TokenList):
     """An 'if' clause with possible 'else if' or 'else' parts."""
-    __slots__ = ('value', 'ttype', 'tokens')
+    M_OPEN = T.Keyword, 'IF'
+    M_CLOSE = T.Keyword, 'END IF'
 
 
 class For(TokenList):
     """A 'FOR' loop."""
-    __slots__ = ('value', 'ttype', 'tokens')
+    M_OPEN = T.Keyword, ('FOR', 'FOREACH')
+    M_CLOSE = T.Keyword, 'END LOOP'
 
 
 class Comparison(TokenList):
     """A comparison used for example in WHERE clauses."""
-    __slots__ = ('value', 'ttype', 'tokens')
 
     @property
     def left(self):
@@ -599,7 +518,6 @@ class Comparison(TokenList):
 
 class Comment(TokenList):
     """A comment."""
-    __slots__ = ('value', 'ttype', 'tokens')
 
     def is_multiline(self):
         return self.tokens and self.tokens[0].ttype == T.Comment.Multiline
@@ -607,15 +525,18 @@ class Comment(TokenList):
 
 class Where(TokenList):
     """A WHERE clause."""
-    __slots__ = ('value', 'ttype', 'tokens')
+    M_OPEN = T.Keyword, 'WHERE'
+    M_CLOSE = T.Keyword, (
+        'ORDER', 'GROUP', 'LIMIT', 'UNION', 'UNION ALL', 'EXCEPT',
+        'HAVING', 'RETURNING', 'INTO')
 
 
 class Case(TokenList):
     """A CASE statement with one or more WHEN and possibly an ELSE part."""
+    M_OPEN = T.Keyword, 'CASE'
+    M_CLOSE = T.Keyword, 'END'
 
-    __slots__ = ('value', 'ttype', 'tokens')
-
-    def get_cases(self):
+    def get_cases(self, skip_ws=False):
         """Returns a list of 2-tuples (condition, value).
 
         If an ELSE exists condition is None.
@@ -631,6 +552,9 @@ class Case(TokenList):
             if token.match(T.Keyword, 'CASE'):
                 continue
 
+            elif skip_ws and token.ttype in T.Whitespace:
+                continue
+
             elif token.match(T.Keyword, 'WHEN'):
                 ret.append(([], []))
                 mode = CONDITION
@@ -663,22 +587,22 @@ class Case(TokenList):
 class Function(TokenList):
     """A function or procedure call."""
 
-    __slots__ = ('value', 'ttype', 'tokens')
-
     def get_parameters(self):
         """Return a list of parameters."""
         parenthesis = self.tokens[-1]
-        for t in parenthesis.tokens:
-            if isinstance(t, IdentifierList):
-                return t.get_identifiers()
-            elif isinstance(t, Identifier) or \
-                isinstance(t, Function) or \
-                t.ttype in T.Literal:
-                return [t,]
+        for token in parenthesis.tokens:
+            if isinstance(token, IdentifierList):
+                return token.get_identifiers()
+            elif imt(token, i=(Function, Identifier), t=T.Literal):
+                return [token, ]
         return []
 
 
 class Begin(TokenList):
     """A BEGIN/END block."""
+    M_OPEN = T.Keyword, 'BEGIN'
+    M_CLOSE = T.Keyword, 'END'
 
-    __slots__ = ('value', 'ttype', 'tokens')
+
+class Operation(TokenList):
+    """Grouping of operations"""
